import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xlux import XLUXExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXLUXCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XLUXExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # XLUX is open from 9:00 am to 5:40 pm.
        yield 8 + (2 / 3)

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2025
            "2025-01-01",  # New Year’s Day
            "2025-04-18",  # Good Friday
            "2025-04-21",  # Easter Monday
            "2025-05-01",  # Labour Day
            "2025-12-25",  # Christmas Day
            "2025-12-26",  # Boxing Day
            # 2024
            "2024-01-01",  # New Year’s Day
            "2024-03-29",  # Good Friday
            "2024-04-01",  # Easter Monday
            "2024-05-01",  # Labour Day
            "2024-12-25",  # Christmas Day
            "2024-12-26",  # Boxing Day
            # 2023
            "2023-01-01",  # New Year’s Day
            "2023-04-07",  # Good Friday
            "2023-04-10",  # Easter Monday
            "2023-05-01",  # Labour Day
            "2023-12-25",  # Christmas Day
            "2023-12-26",  # Boxing Day
            # 2022
            "2022-01-01",  # New Year’s Day
            "2022-04-15",  # Good Friday
            "2022-04-18",  # Easter Monday
            "2022-05-01",  # Labour Day
            "2022-12-25",  # Christmas Day
            "2022-12-26",  # Boxing Day
            # 2021
            "2021-01-01",  # New Year’s Day
            "2021-04-02",  # Good Friday
            "2021-04-05",  # Easter Monday
            "2021-05-01",  # Labour Day
            "2021-12-25",  # Christmas Day
            "2021-12-26",  # Boxing Day
            # 2020
            "2020-01-01",  # New Year’s Day
            "2020-04-10",  # Good Friday
            "2020-04-13",  # Easter Monday
            "2020-05-01",  # Labour Day
            "2020-12-25",  # Christmas Day
            "2020-12-26",  # Boxing Day
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            "2025-01-02",
            "2025-01-14",
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            # 2025
            "2025-12-24",  # Christmas Eve
            "2025-12-31",  # New Year's Eve
            # 2024
            "2024-12-24",  # Christmas Eve
            "2024-12-31",  # New Year's Eve
            # 2021
            "2021-12-24",  # Christmas Eve
            "2021-12-31",  # New Year's Eve
            # 2020
            "2020-12-24",  # Christmas Eve
            "2020-12-31",  # New Year's Eve
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(hours=14, minutes=5)



import pytest

from exchange_calendars.exchange_calendar_xlon import XLONExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXLONCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XLONExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 8.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2012
            "2012-01-02",  # New Year's observed
            "2012-04-06",  # Good Friday
            "2012-04-09",  # Easter Monday
            "2012-05-07",  # May Day
            "2012-06-04",  # Spring Bank Holiday
            "2012-08-27",  # Summer Bank Holiday
            "2012-12-25",  # Christmas
            "2012-12-26",  # Boxing Day
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            "2002-06-03",  # Spring Bank 2002
            "2002-06-04",  # Golden Jubilee
            "2011-04-29",  # Royal Wedding
            "2012-06-04",  # Spring Bank 2012
            "2012-06-05",  # Diamond Jubilee
            "2020-05-08",  # VE Day
            "2022-09-19",  # Queen Elizabeth Funeral
            "2023-05-08",  # Coronation of King Charles III
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        # May Bank Holiday was instead observed on VE Day in 2020.
        yield ["2020-05-04"]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            "2012-12-24",  # Christmas Eve
            "2012-12-31",  # New Year's Eve
            "2010-12-24",  # Christmas Eve
            "2010-12-31",  # New Year's Eve
            # In Dec 2011, Christmas Eve and NYE are both on a Saturday,
            # so preceding Fridays (the 23rd and 30th) are early closes.
            "2011-12-23",
            "2011-12-30",
        ]



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xkls import XKLSExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXKLSCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XKLSExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XKLS is open from 9AM to 5PM
        yield 8

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2019-01-01",  # New Year's Day
            "2019-01-21",  # Thaipusam
            "2019-02-01",  # Federal Territory Day
            "2019-02-05",  # Chinese New Year Day 1
            "2019-02-06",  # Chinese New Year Day 2
            "2019-05-01",  # Labour Day
            "2019-05-20",  # Wesak Day
            "2019-05-22",  # Nuzul Al Quran
            "2019-06-05",  # Eid al Fitr Day 1
            "2019-06-06",  # Eid al Fitr Day 2
            "2019-08-12",  # Eid al Adha
            "2018-08-31",  # National Day
            "2019-09-02",  # Muharram
            "2019-09-16",  # Malaysia Day
            "2019-10-28",  # Deepavali
            "2018-11-20",  # Birthday of Muhammad
            "2019-12-25",  # Christmas Day
            #
            # Holidays falling on Sundays are made up on the next Monday.
            "2012-01-02",  # New Year's Day
            "2015-02-02",  # Federal Territory Day
            "2017-01-30",  # Chinese New Year 2nd Day
            "2016-05-02",  # Labour Day
            "2014-09-01",  # National Day
            "2012-09-17",  # Malaysia Day
            "2016-12-26",  # Christmas Day
            "2020-05-25",  # Eid al Fitr
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield []

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays that fall on a weekend are not made up. Ensure surrounding
            # days are not holidays.
            # New Year's Day on Saturday, Jan 1st.
            #   2010-12-31 is an adhoc holiday
            "2010-12-30",
            "2011-01-03",
            # Labour Day on Saturday, May 1st.
            "2010-04-30",
            "2010-05-03",
            # Nuzul Al-Quran on Saturday, Jul 4th.
            "2017-07-03",
            "2017-07-06",
            # National Day on Saturday, Aug 31st.
            #   2019-09-02 is Muharram
            "2019-08-30",
            "2019-09-03",
            # Muharram on Saturday, Oct 25th.
            "2014-10-24",
            "2014-10-27",
            # Malaysia Day on Saturday, Sep 16th.
            "2017-09-15",
            "2017-09-18",
            # Birthday of Prophet Muhammad on Saturday, Mar 31st.
            "2007-03-30",
            "2007-04-02",
            # Christmas Day on Saturday, Dec 25th.
            "2010-12-24",
            "2010-12-27",
            # Thaipusam on Saturday, Feb 8th.
            "2020-02-07",
            "2020-02-10",
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            "2019-02-04",  # Day before Chinese New Year.
            "2019-06-04",  # Day before Eid al Fitr.
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(hours=12, minutes=30)



import pytest

from exchange_calendars.exchange_calendar_xpra import XPRAExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXPRACalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XPRAExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XPRA is open from 9:00 to 4:20PM
        yield 7.34

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2019 +
            "2019-01-01",  # New Year's Day
            "2019-04-19",  # Good Friday
            "2019-04-22",  # Easter Monday
            "2019-05-01",  # Labour Day
            "2019-05-08",  # Liberation Day
            "2019-07-05",  # St. Cyril/Methodius Day
            "2018-07-06",  # Jan Hus Day
            "2018-09-28",  # Czech Statehood Day
            "2019-10-28",  # Independence Day
            "2017-11-17",  # Freedom/Democracy Day
            "2019-12-24",  # Christmas Eve
            "2019-12-25",  # Christmas Day
            "2019-12-26",  # Second Day of Christmas
            "2019-12-31",  # Exchange Holiday
            #
            # Good Friday is a regular holiday from 2013
            "2013-03-29",
            "2014-04-18",
            "2015-04-03",
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            "2002-08-14",  # Extreme Flooding
            "2004-01-02",  # Restoration of the Czech Independence Day
            "2005-01-03",  # Restoration of the Czech Independence Day
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays that fall on a weekend and are not made up. Ensure surrounding
            # days are not holidays.
            # New Years Day on Sunday, Jan 1st
            "2011-12-30",
            "2012-01-02",
            # Labour Day on Sunday, May 1st
            "2016-04-29",
            "2016-05-02",
            # Liberation Day on Sunday, May 8th
            "2016-05-06",
            "2016-05-09",
            # Saints Cyril and Methodius Day on Saturday, Jul 5th
            "2014-07-04",
            "2014-07-07",
            # Jan Hus Day on Saturday, Jul 6th
            #   Note: 7/5/2019 is a holiday
            "2019-07-04",
            "2019-07-08",
            # Czech Statehood Day on Saturday, Sept 28th
            "2019-09-27",
            "2019-09-30",
            # Independence Day on Sunday, Oct 28th
            "2018-10-26",
            "2018-10-29",
            # Struggle for Freedom and Democracy Day on Sunday, Nov 17
            "2019-11-15",
            "2019-11-18",
            # Christmas Eve on a Sunday
            #   Note: 25th, 26th both holidays
            "2017-12-22",
            "2017-12-27",
            # Christmas on a Sunday
            #   Note: 26th a holiday
            "2016-12-23",
            "2016-12-27",
            # 2nd Day of Christmas on Saturday, Dec 26
            #   Note: 25th, 24th both holidays
            "2015-12-23",
            "2015-12-28",
            # Exchange Holiday on Saturday, Dec 31
            "2016-12-30",
            "2017-01-02",
            #
            # Good Friday was not a holiday prior to 2013
            "2010-04-02",
            "2011-04-22",
            "2012-04-06",
        ]



import pytest

from exchange_calendars.exchange_calendar_xbse import XBSEExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXBSECalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XBSEExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # XBSE is open from 10:00 to 5:20PM on its longest trading day
        yield 7 + (3 / 4)

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2023-04-17",  # Orthodox Easter
            "2023-04-14",  # Good Friday
            "2021-01-01",  # New Year's Day
            "2021-04-30",  # Orthodox Good Friday
            "2021-06-01",  # Children's day
            "2021-06-21",  # Orthodox Pentecost
            "2021-11-30",  # St. Adnrew's Day
            "2021-12-01",  # National Day
            "2020-01-01",  # New Year's Day
            "2020-01-02",  # New Year's Day
            "2020-01-24",  # Romanian Principalities Unification Day
            "2020-04-17",  # Good Friday
            "2020-04-20",  # Orthodox Easter
            "2020-05-01",  # Labour Day
            "2020-06-01",  # Children's Day
            "2020-06-08",  # Orthodox Pentecost
            "2020-11-30",  # St. Adnrew's day
            "2020-12-01",  # National Day
            "2020-12-25",  # Christmans
            "2019-01-01",  # New Year's Day
            "2019-01-02",  # New Year's Day
            "2019-01-24",  # Romanian Principalities Unification Day
            "2019-04-26",  # Good Friday
            "2019-04-29",  # Orthodox Easter
            "2019-05-01",  # Labour Day
            "2019-06-17",  # Orthodox Pentecost
            "2019-08-15",  # Assumption of Virgin Mary
            "2019-12-25",  # Christmans
            "2019-12-26",  # Christmans
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            # Athens Stock Exchange observes Orthodox (or Eastern) Easter,
            # as well as Western Easter. All holidays that are tethered to
            # Easter (i.e. Whit Monday, Good Friday, etc.), are relative to
            # Orthodox Easter. Following checks that Orthodox Easter and
            # related holidays are correct.
            # # Some Orthodox Good Friday dates
            "2002-05-03",
            "2005-04-29",
            "2008-04-25",
            "2009-04-17",
            "2016-04-29",
            "2017-04-14",
            # Some Orthodox Pentecost dates
            "2002-06-24",
            "2005-06-20",
            "2006-06-12",
            "2008-06-16",
            "2013-06-24",
            "2016-06-20",
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays that fall on a weekend are not made up. Ensure surrounding
            # days are not holidays.
            # Second New Years Day on Saturday, Jan 2st
            "2021-01-04",
            # Christmas 25th, 26th both holidays, fall Saturday Sunday
            "2021-12-24",
            "2021-12-27",
            # Labour Day on Saturday + Good Friday on Friday + Orthodox Easter on Monday
            "2021-04-29",
            "2021-05-04",
            # Children's Day on Saturday
            "2019-05-31",
            "2019-06-03",
            # Assumption of Virgin Mary on Sunday
            "2021-08-13",
            "2021-08-16",
            # Assumption of Virgin Mary on Saturday
            "2020-08-14",
            "2020-08-17",
        ]



import pytest

from exchange_calendars.exchange_calendar_xswx import XSWXExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestIXSWXCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XSWXExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XSWX is open from 9:00 am to 5:30 pm.
        yield 8.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2012
            # New Year's Day isn't observed because it was on a Sunday
            "2012-01-02",  # Berchtold's Day observed
            "2012-04-06",  # Good Friday
            "2012-04-09",  # Easter Monday
            "2012-05-01",  # Labour Day
            "2012-05-17",  # Ascension Day
            "2012-05-28",  # Whit Monday
            "2012-08-01",  # Swiss National Day
            "2012-12-24",  # Christmas Eve
            "2012-12-25",  # Christmas
            "2012-12-26",  # Boxing Day
            "2012-12-31",  # New Year's Eve
        ]



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_iepa import IEPAExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestIEPACalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield IEPAExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 22

    @pytest.fixture
    def regular_holidays_sample(self):
        # new year's, good friday, christmas
        yield ["2016-01-01", "2016-03-25", "2016-12-26"]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield ["2012-10-29"]  # hurricane sandy (day one)

    @pytest.fixture
    def non_holidays_sample(self):
        yield ["2012-10-30"]  # hurricane sandy day two - exchange open

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            "2016-01-18",  # mlk
            "2016-02-15",  # presidents
            "2016-05-30",  # mem day
            "2016-07-04",  # independence day
            "2016-09-05",  # labor
            "2016-11-24",  # thanksgiving
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(13, "h")



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xsto import XSTOExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXSTOCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XSTOExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XSTO is open from 9:00 am to 5:30 pm.
        yield 8.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2018 +
            "2018-01-01",  # New Year's Day
            "2017-01-06",  # Epiphany
            "2018-03-30",  # Good Friday
            "2018-04-02",  # Easter Monday
            "2018-05-01",  # Labour Day
            "2018-05-10",  # Ascension Day
            "2018-06-06",  # National Day
            "2018-06-22",  # Midsummer Eve
            "2018-12-24",  # Christmas Eve
            "2018-12-25",  # Christmas Day
            "2018-12-26",  # Boxing Day
            "2018-12-31",  # New Year's Eve
            #
            # Midsummer Eve falls on the Friday after June 18.
            "2010-06-25",  # 18th a Friday, holiday observed Friday 25th (not 18th).
            "2017-06-23",  # 18th a Sunday, holiday observed following Friday.
            # Holidays that ceased to be observed.
            "2004-05-31",  # Whit Monday last observed in 2004.
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Midsummer Eve falls on the Friday after June 18.
            "2010-06-18",  # 18th a Friday, holiday observed Friday 25th (not 18th)
            #
            # Holidays that fall on a weekend and are not made up. Ensure surrounding
            # days are not holidays.
            # In 2018, the Epiphany fell on a Saturday, so the market should be
            # open on both the prior Friday and the following Monday.
            "2018-01-05",
            "2018-01-08",
            # In 2010, Labour Day fell on a Saturday, so the market should be
            # open on both the prior Friday and the following Monday.
            "2010-04-30",
            "2010-05-03",
            # In 2015, National Day fell on a Saturday, so the market should be
            # open on both the prior Friday and the following Monday.
            "2015-06-05",
            "2015-06-08",
            # In 2010, Christmas fell on a Saturday, meaning Boxing Day fell on
            # a Sunday. The market should thus be open on the following Monday.
            "2010-12-27",
            # In 2017, New Year's Day fell on a Sunday, so the market should be
            # open on both the prior Friday and the following Monday.
            "2016-12-30",
            "2017-01-02",
            #
            # Holidays that ceased to be observed.
            "2005-05-16",  # Whit Monday ceased to be observed from 2005.
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            "2018-01-05",  # Day before Epiphany.
            "2018-03-29",  # Maundy Thursday.
            "2018-04-30",  # Day before Labour Day.
            "2018-05-09",  # Day before Ascension Day.
            "2018-11-02",  # All Saints' Eve.
            "2015-10-30",  # All Saints' Eve.
            "2010-11-05",  # All Saints' Eve.
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(13, "h")



from datetime import time
from zoneinfo import ZoneInfo

from pandas.tseries.holiday import EasterMonday, GoodFriday

from .common_holidays import (
    boxing_day,
    christmas,
    christmas_eve,
    european_labour_day,
    new_years_day,
    new_years_eve,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()
LabourDay = european_labour_day()
ChristmasEve = christmas_eve()
Christmas = christmas()
BoxingDay = boxing_day()
NewYearsEve = new_years_eve()


class XEEEExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the European Energy Exchange AG, Leipzig, Germany.
    https://www.eex.com/fileadmin/EEX/Downloads/Trading/Calendar/Holiday_Calendar/20230303_Trading_Calendar_EEX_Group.pdf

    Open Time: 8:00 AM, CET (Central European Time)
    Close Time: 6:00 PM, CET (Central European Time)

    Regularly-Observed Holidays:
      - New Year's Day
      - Good Friday
      - Easter Monday
      - Labour Day
      - Christmas Eve
      - Christmas Day
      - Boxing Day
      - New Year's Eve
    """

    name = "XEEE"

    tz = ZoneInfo("Europe/Berlin")

    open_times = ((None, time(8)),)

    close_times = ((None, time(18, 0)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                GoodFriday,
                EasterMonday,
                LabourDay,
                ChristmasEve,
                Christmas,
                BoxingDay,
                NewYearsEve,
            ]
        )



from datetime import time
from itertools import chain
from zoneinfo import ZoneInfo

import pandas as pd
from pandas.tseries.holiday import Holiday
import numpy as np  # for concatenate

from exchange_calendars.exchange_calendar import ExchangeCalendar, HolidayCalendar

SaudiFoundingDay = Holiday("Saudi Founding Day", month=2, day=22)

NationalDayOfSaudiArabia = Holiday("National Day of Saudi Arabia", month=9, day=23)

# https://www.saudiexchange.sa/wps/portal/saudiexchange/about-saudi-exchange/exchange-media-centre/saudi-exchange-holiday-calendar?locale=en
EidAlAdhaHoliday = pd.to_datetime(
    np.concatenate(
        [
            pd.date_range("2025-06-04", "2025-06-09"),
            pd.date_range("2024-06-13", "2024-06-23"),
            pd.date_range("2023-06-27", "2023-07-02"),
            pd.date_range("2022-07-07", "2022-07-13"),
            pd.date_range("2021-07-16", "2021-07-22"),
        ]
    )
)

EidAlFiterHoliday = pd.to_datetime(
    np.concatenate(
        [
            pd.date_range("2025-03-27", "2024-04-02"),
            pd.date_range("2024-04-05", "2024-04-15"),
            pd.date_range("2023-04-18", "2023-04-25"),
            pd.date_range("2022-04-28", "2022-05-08"),
            pd.date_range("2021-05-13", "2021-05-16"),
        ]
    )
)


class XSAUExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Saudi Exchange (XSAU)
    Available here: https://www.saudiexchange.sa/wps/portal/saudiexchange/rules-guidance/capital-market-overview/trading-cycle-and-times
    """

    name = "XSAU"

    tz = ZoneInfo("Asia/Riyadh")

    open_times = ((None, time(10)),)

    close_times = ((None, time(15)),)

    @classmethod
    def bound_min(cls) -> pd.Timestamp:
        return pd.Timestamp("2021-01-01")

    @classmethod
    def bound_max(cls) -> pd.Timestamp:
        return pd.Timestamp("2025-12-31")

    @property
    def regular_holidays(self):
        return HolidayCalendar([SaudiFoundingDay, NationalDayOfSaudiArabia])

    @property
    def weekmask(self):
        return "1111001"

    @property
    def adhoc_holidays(self):
        return list(chain(EidAlAdhaHoliday, EidAlFiterHoliday))



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xist import XISTExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXISTCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XISTExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XIST is open from 10:00 am to 6:00 pm
        yield 8

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2019-01-01",  # New Year's Day
            "2019-04-23",  # Natl Sov and Children's Day
            "2019-05-01",  # Labour Day
            "2017-05-19",  # CAYS Day
            "2019-06-04",  # Eid al Fitr Day 1
            "2019-06-05",  # Eid al Fitr Day 2
            "2019-06-06",  # Eid al Fitr Day 3
            "2019-07-15",  # Dem and Natl Unity Day
            "2016-09-12",  # Eid al Adha Day 1
            "2016-09-13",  # Eid al Adha Day 2
            "2016-09-14",  # Eid al Adha Day 3
            "2016-09-15",  # Eid al Adha Day 4
            "2019-08-30",  # Victory Day
            "2019-10-29",  # Republic Day
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            # Miscellaneous closures
            "2002-01-04",  # Market Holiday
            "2003-11-21",  # Terror attacks
            "2004-01-23",  # Bad weather
            "2004-12-30",  # Closure for redenomination
            "2004-12-31",  # Closure for redenomination
            # Eid al Adha and Eid al Fitr extra closures
            "2003-02-14",  # Eid al Adha extra holiday
            "2003-11-24",  # Eid al Fitr extra holiday
            "2003-11-28",  # Eid al Fitr extra holiday
            "2006-01-13",  # Eid al Adha extra holiday
            "2023-02-08",  # Trade suspension after earthquake
            "2023-02-09",  # Trade suspension after earthquake
            "2023-02-10",  # Trade suspension after earthquake
            "2023-02-13",  # Trade suspension after earthquake
            "2023-02-14",  # Trade suspension after earthquake
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays that fall on a weekend are not made up. Ensure surrounding
            # days are not holidays.
            # New Year's Day on Sunday, Jan 1st.
            "2011-12-30",
            "2012-01-02",
            # Natl Sovereignty and Children's Day on Sunday, Apr 23rd.
            "2017-04-21",
            "2017-04-24",
            # Labour Day on Sunday, May 1st.
            "2016-04-29",
            "2016-05-02",
            # Com. of Attaturk Youth and Sport's Day on Saturday, May 19th.
            "2018-05-18",
            "2018-05-21",
            # Eid Al Fitr (Day 3) on Sunday, Jun 17th (Friday is a holiday).
            "2018-06-18",
            # Democracy and National Unity Day on Sunday, Jul 15th.
            "2018-08-13",
            "2018-07-16",
            # Eid Al Adha (Day 1) on Sunday, Aug 11th (Monday is a holiday).
            "2019-08-09",
            # Victory Day on Saturday, Aug 30th.
            "2014-08-29",
            "2014-09-01",
            # Republic Day on Saturday, Oct 29th.
            "2016-10-28",
            "2016-10-31",
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            "2019-10-28",  # Day before Republic Day.
            "2019-06-03",  # Day before Eid al Adha.
            "2018-08-20",  # Day before Eid al Adha.
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(hours=12, minutes=30)



"""Classes and functions to explore the bounds of calendar factories.

Jul 21. Module written (prior to implementation of `bound_min`,
`bound_max`) to explore the bounds of calendar factories. Provides for
evaluating the earliest start date and latest end date for which a
calendar can be instantiated without raising an error. Also records errors
raised when dates are passed beyond these limits.

Module retained in case might be useful.
"""

import abc
import dataclasses
import pathlib
import pickle
from typing import Literal, Type

import pandas as pd

import exchange_calendars as xcals
from exchange_calendars.calendar_helpers import UTC


@dataclasses.dataclass
class FactoryBounds:
    """Bounds within which an xcals.ExchangeCalendar can be calculated.

    Parameters
    ----------
    name :
        Name of calendar with declared bounds.

    start : pd.Timestamp
        Earliest start date for which can create a calendar with end date
        as tomorrow.

    start_error :
        Error instance raised in event request a calendar with start date
        one day earlier than `start` and end date as tomorrow.

    end : pd.Timestamp
        Latest end date for which can create a calendar with start date
        as yesterday.

    end_error :
        Error instance raised in event request a calendar with end date
        one day later than `end` and start date as yesterday.
    """

    name: str
    start: pd.Timestamp
    start_error: Exception
    end: pd.Timestamp
    end_error: Exception


class _FindFactoryBounds:
    """Find start and end bounds of a given calendar factory."""

    def __init__(self, Factory: xcals.ExchangeCalendar, watch: bool = False):
        self.Factory = Factory
        self.watch = watch
        self._last_error: Exception | None

    @property
    def calendar_name(self) -> str:
        """Calendar name."""
        if isinstance(self.Factory.name, str):
            return self.Factory.name
        else:
            return self.Factory().name

    @property
    def today(self) -> pd.Timestamp:
        return pd.Timestamp.now(tz=UTC).floor("D")

    def _get_calendar(
        self,
        start: pd.Timestamp | None = None,
        end: pd.Timestamp | None = None,
    ) -> xcals.ExchangeCalendar:
        """Get calendar for period between now and `start` or `end`."""
        if self.watch:
            insert = f"start={start}" if start is not None else f"end={end}"
        if start is None:
            start = self.today - pd.Timedelta(1, "D")
        elif end is None:
            end = self.today + pd.Timedelta(1, "D")
        else:
            raise ValueError("`start` and `end` cannot both be None.")
        if self.watch:
            print(f"getting calendar '{self.calendar_name}' with {insert}.")
        return self.Factory(start=start, end=end)

    def _is_valid_date(
        self, date: pd.Timestamp, bound: Literal["start", "end"]
    ) -> bool:
        kwargs = {bound: date}
        try:
            self._get_calendar(**kwargs)
        except Exception as err:  # pylint: disable=broad-except
            self._last_error = err
            return False
        else:
            return True

    def _get_a_valid_date_by_trying_every_x_days(
        self,
        look_from: pd.Timestamp,
        offset: pd.DateOffset,
        bound: Literal["start", "end"],
    ) -> pd.Timestamp:
        # recursively look for a valid date every offset, return first that's valid
        if self._is_valid_date(look_from, bound):
            return look_from
        else:
            next_look_from = look_from + offset
            if (  # if start has move into the future or end into the past
                look_from <= self.today < next_look_from
                or look_from >= self.today > next_look_from
            ):
                return self.today
            else:
                return self._get_a_valid_date_by_trying_every_x_days(
                    next_look_from, offset, bound
                )

    @property
    def _first_offset(self) -> pd.DateOffset:
        return pd.DateOffset(years=100)

    def _is_first_offset(self, offset: pd.DateOffset) -> bool:
        return offset == self._first_offset or -offset == self._first_offset

    def _offset_iterator(
        self, bound: Literal["start", "end"]
    ) -> abc.Iterator[pd.DateOffset]:
        sign = 1 if bound == "start" else -1
        iterator = iter(
            [
                sign * self._first_offset,
                sign * pd.DateOffset(years=30),
                sign * pd.DateOffset(years=10),
                sign * pd.DateOffset(years=3),
                sign * pd.DateOffset(years=1),
                sign * pd.DateOffset(months=3),
                sign * pd.DateOffset(months=1),
                sign * pd.DateOffset(days=10),
                sign * pd.DateOffset(days=3),
                sign * pd.DateOffset(days=1),
            ]
        )
        return iterator

    def _is_valid_bound(
        self, date: pd.Timestamp, bound: Literal["start", "end"]
    ) -> bool:
        if not self._is_valid_date(date, bound):
            return False
        else:
            day_delta = 1 if bound == "end" else -1
            date = date + pd.Timedelta(day_delta, "D")
            return not self._is_valid_date(date, bound)

    def _try_short_cut(
        self, bound: Literal["start", "end"]
    ) -> tuple[pd.Timestamp, Exception] | None:
        """Try known likely bounds around min/max Timestamp.

        These likely bounds are caused by how special closes is calculated.

        Return None if no likely bound is a bound.
        """
        if bound == "start":
            likely_bounds = [
                pd.Timestamp("1678-01-01", tz=UTC),
                pd.Timestamp("1679-01-01", tz=UTC),
            ]
        else:
            likely_bounds = [
                pd.Timestamp("2260-12-31", tz=UTC),
                pd.Timestamp("2262-04-10", tz=UTC),
            ]

        for likely_bound in likely_bounds:
            if self._is_valid_bound(likely_bound, bound):
                assert self._last_error is not None
                return (likely_bound, self._last_error)
        return None

    @staticmethod
    def _initial_value(bound: Literal["start", "end"]) -> pd.Timestamp:
        if bound == "start":
            return pd.Timestamp.min.ceil("D").tz_localize(UTC)
        else:
            return pd.Timestamp.max.floor("D").tz_localize(UTC)

    def _get_bound(
        self, bound: Literal["start", "end"]
    ) -> tuple[pd.Timestamp, Exception]:
        self._last_error = None
        initial_value = self._initial_value(bound)

        try:
            assert not self._is_valid_bound(initial_value, bound)
        except pd.errors.OutOfBoundsDatetime as err:
            # if initial_value is a valid bound then `is_valid_bound` will
            # raise OutOfBoundsDatetime as prod beyond pandas timestamp bounds
            return (initial_value, err)

        offset_iterator = self._offset_iterator(bound)
        offset = next(offset_iterator)
        look_from = initial_value
        while offset:
            look_from += (
                offset  # avoid getting calendar with same parameters as a previous call
            )
            valid_date = self._get_a_valid_date_by_trying_every_x_days(
                look_from, offset, bound
            )
            if valid_date == look_from and self._is_first_offset(offset):
                # first attempt was valid, chances are only calculation of
                # special closes was causing error on min date. Try short-cut:
                short_cut = self._try_short_cut(bound)
                if short_cut is not None:
                    return short_cut
            look_from = valid_date - offset
            try:
                offset = next(offset_iterator)
            except StopIteration:
                break
        return (valid_date, self._last_error)

    def bounds(self) -> FactoryBounds:
        """Return valid bounds for Factory's `start` and `end` parameters.

        See find_factory_bounds.___doc___
        """
        start_bound = self._get_bound("start")
        end_bound = self._get_bound("end")
        return FactoryBounds(self.calendar_name, *start_bound, *end_bound)


def find_factory_bounds(
    Factory: xcals.ExchangeCalendar, watch: bool = False
) -> FactoryBounds:
    """Return FactoryBounds for a factory.

    Bounds deliniate the longest continuous period, which includes today,
    through which start and end dates will always return an
    xcals.ExchangeCalendar. Calling the `Factory` with a start date one day
    before the start bound, or with an end date one day after the end
    bound, will raise an error.

    Parameters
    ----------
    Factory :
        xcals.ExchangeCalendar class to query bounds of.

    watch :
        True for a live print of factory calls as the bounds are
        sought out and prodded.

    Returns
    -------
    FactoryBounds:
        Dataclass that describes bounds and associated out-of-bounds
        errors. See FactoryBounds.__doc__.

    Notes
    -----
    Uses trial and error to home in on and prod the bounds.
    """
    return _FindFactoryBounds(Factory, watch).bounds()


def _find_bounds_all_factories(watch=False) -> dict[str, FactoryBounds]:
    """Find and return Factory Bounds for all calendars.

    NOTE: This function will take a fair while to run! Pass `watch` as
    True for a live print of factory calls as the bounds are sought and
    proded.
    """
    cal_bounds = {}
    for name, Factory in xcals.calendar_utils._default_calendar_factories.items():
        cal_bounds[name] = find_factory_bounds(Factory, watch)
    return cal_bounds


def _factory_bounds_resource_path() -> pathlib.PurePath:
    path = pathlib.Path(__file__)
    path = pathlib.Path(path.parent, "resources/_factory_bounds.dat")
    assert path.exists()
    return path


def _bake_all_calendar_bounds(all_calendar_bounds: dict["str", FactoryBounds]):
    """Bake `all_calendar_bounds` to resource file.

    NOTE: CHECK passed `all_calendar_bounds` before baking.

    Parameters
    ----------
    all_calendar_bounds :
        As returned by `_find_bounds_all_factories`. CHECK return before
        baking!!
    """
    assert isinstance(all_calendar_bounds, dict)
    assert len(all_calendar_bounds) > 20
    key = list(all_calendar_bounds.keys())[0]
    assert isinstance(key, str)
    assert isinstance(all_calendar_bounds[key], FactoryBounds)
    with open(_factory_bounds_resource_path(), "wb") as file:
        pickle.dump(all_calendar_bounds, file)


def _retrieve_all_calendars_bounds() -> dict["str", FactoryBounds]:
    """Retrieve `all_calendar_bounds` from resources."""
    with open(_factory_bounds_resource_path(), "rb") as file:
        all_calendar_bounds = pickle.load(file)
    return all_calendar_bounds


_all_calendars_bounds = _retrieve_all_calendars_bounds()


def _get_all_calendars_bounds_exceptions() -> tuple[Type[Exception], ...]:
    """Exceptions the can be raised when request calendar with oob date."""
    bounds = _all_calendars_bounds
    errors = []
    for cal in bounds:
        error_types = (type(bounds[cal].start_error), type(bounds[cal].end_error))
        for error_type in error_types:
            if error_type not in errors:
                errors.append(error_type)
    return tuple(errors)


def all_calendars_bounds() -> pd.DataFrame:
    """Return DataFrame describing start and end bounds for each calendar.

    Bounds deliniate the longest continuous period, which includes today,
    through which start and end dates will always return an
    xcals.ExchangeCalendar.

    Returns
    -------
    pd.DataFrame
        Indexed with calendar names.
        Columns:
            earliest_valid_start - earliest start date from which a
                calendar can be instantiated with end date as today.
            latest_valid_end - latest end date to which a calendar can
                be instantiated with start date as today.
    """
    cal_bounds = _all_calendars_bounds
    start_bounds = []
    end_bounds = []
    for _, factory_bounds in cal_bounds.items():
        start_bounds.append(factory_bounds.start)
        end_bounds.append(factory_bounds.end)

    return pd.DataFrame(
        dict(earliest_valid_start=start_bounds, latest_valid_end=end_bounds),
        index=cal_bounds.keys(),
    )


def get_calendar_bounds(
    calendar_name: str,
) -> tuple[pd.Timestamp, pd.Timestamp]:
    """Retun 2-tuple describing start and end bounds for a given calendar.

    Bounds deliniate the longest continuous period, which includes today,
    through which start and end dates will always return a calendar.
    """
    bounds = _all_calendars_bounds[calendar_name]
    return (bounds.start, bounds.end)



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xbud import XBUDExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXBUDCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XBUDExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XBUD is open from 9:00 to 5:00PM
        yield 8

    @pytest.fixture
    def regular_holidays_sample(self):
        hols = [
            # 2019
            "2019-01-01",  # New Year's Day
            "2019-03-15",  # National Holiday
            "2019-04-19",  # Good Friday
            "2019-04-22",  # Easter Monday
            "2019-05-01",  # Labour Day
            "2019-06-10",  # Whit Monday
            "2019-08-20",  # St. Stephen's Day
            "2019-10-23",  # National Holiday
            "2019-11-01",  # All Saint's Day
            "2019-12-24",  # Christmas Eve
            "2019-12-25",  # Christmas Day
            "2019-12-26",  # Second Day of Christmas
            "2019-12-31",  # New Year's Eve
            # Bridge Days. If some holidays fall on a Tuesday or Thursday
            # then Monday or Friday is also a holiday (bridge day).
            # New Years Day on Thursday, Jan 1
            "2015-01-01",
            "2015-01-02",
            # National Holiday on Tuesday, March 15
            "2016-03-15",
            "2016-03-14",
            # Labour Day on Tuesday, May 1
            "2018-05-01",
            "2018-04-30",
            # St. Stephen's Day on Thursday, Aug 20
            "2015-08-20",
            "2015-08-21",
            # National Holiday on Thursday, Oct 23
            "2014-10-23",
            "2014-10-24",
            # All Saint's Day on Tuesday, Nov 1
            "2016-11-01",
            "2016-10-31",
            # Second Day of Christmas on Thursday, Dec 26
            "2019-12-27",
            # Observance of bridge day when Second Day of Christmas falls on a Thursday
            # is not strict, although was observed in 2013, 2019.
            "2013-12-27",  # Friday
            "2019-12-27",  # Friday
        ]
        # New Year's Eve only a holiday from 2011.
        hols.extend([f"{year}-12-31" for year in range(2011, 2020)])
        yield hols

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield []

    @pytest.fixture
    def non_holidays_sample(self):
        non_hols = [
            # Holidays that fall on a weekend are not made up. Ensure surrounding
            # days are not holidays.
            # New Years Day on Sunday, Jan 1st
            "2011-12-30",
            "2012-01-02",
            # National Holiday on Sunday, March 15th
            "2015-03-13",
            "2015-03-16",
            # Labour Day on Sunday, May 1st
            "2016-04-29",
            "2016-05-02",
            # St. Stephen's Day on Saturday, August 20th
            "2016-08-19",
            "2016-08-22",
            # National Holiday on Sunday, Oct 23rd
            "2016-10-21",
            "2016-10-24",
            # All Saint's Day on Sunday, Nov 1
            "2015-10-30",
            "2015-11-02",
            # Christmas Eve on a Sunday. Note: 25th, 26th both holidays
            "2017-12-22",
            "2017-12-27",
            # Christmas on a Sunday. Note: 26th a holiday
            "2016-12-23",
            "2016-12-27",
            # 2nd Day of Christmas on Saturday, Dec 26. Note: 25th, 24th both holidays
            "2015-12-23",
            "2015-12-28",
            # New Year's Eve on Saturday, Dec 31
            "2016-12-30",
            "2017-01-02",
            #
            # Christmas Eve and New Years Eve are holidays although do not provoke
            # a bridge day if they fall on a Tuesday or Thursday. Following ensures
            # that what would be the bridge day is not a holiday.
            # Dec 23 Mondays
            "2002-12-23",
            "2013-12-23",
            "2019-12-23",
            # Dec 30 Mondays
            "2002-12-30",
            "2013-12-30",
            "2019-12-30",
            #
            # Although strictly a bridge day, 2002-12-27 is not observed as a holiday
            # (falls on a Friday following a holiday (Second Day of Christmas)).
            "2002-12-27",
        ]
        # Prior to 2011 New Year's Eve was only a holiday if it fell on a Monday (in
        # which case was treated as a bridge day). Check that nye not a holiday prior
        # to 2011 if it didn't fall on a Monday (or weekend).
        nye = [f"{year}-12-31" for year in range(2000, 2011)]
        nye_sessions = [
            date for date in nye if pd.Timestamp(date).weekday() not in [0, 5, 6]
        ]
        non_hols.extend(nye_sessions)
        yield non_hols



"""
This script can be used to generate the CSV files used in the exchange calendar
tests. The CSVs include a calendar's sessions, open times, close times,
break_start times and break_end times.

This script can be run from the root of the repository with:
    $ python etc/make_exchange_calendar_test_csv.py <calendar_iso_code> <start> <end>
where:
    <calendar_iso_code>
        Code for the corresponding calendar, e.g. "XNYS"

    <start> optional, default: "existing" if existing, else "default"
        First session to include to CSV file. Can take any of:
            A date, for example "1998-06-15".
            "existing" to use first session of any exising csv file.
            "default" to use calendar's default start date.

    <end> optional, default: "default"
        NB if <end> is provided then <start> MUST also be provided.
        Last session to include to CSV file. Can take any of:
            A date, for example "2026-06-15".
            "existing" to use last session of any exising csv file.
            "default" to use calendar's default end date.
"""

import sys
import pathlib

import pandas as pd

from exchange_calendars import get_calendar

cal_name = sys.argv[1].upper()

filename = cal_name.replace("/", "-").lower() + ".csv"
path = pathlib.Path(__file__).parents[1].joinpath("tests/resources", filename)

if len(sys.argv) > 2:
    start_arg = sys.argv[2]
else:
    start_arg = "existing" if path.is_file else "default"
end_arg = sys.argv[3] if len(sys.argv) > 3 else "default"

if start_arg == "existing" or end_arg == "existing":
    assert path.is_file
    df = pd.read_csv(
        path,
        index_col=0,
        parse_dates=[0, 1, 2, 3, 4],
        infer_datetime_format=True,
    )

if start_arg == "existing":
    start = df.index[0].tz_localize(None)
elif start_arg == "default":
    start = None
else:
    start = start_arg

if end_arg == "existing":
    end = df.index[-1].tz_localize(None)
elif end_arg == "default":
    end = None
else:
    end = end_arg

cal = get_calendar(cal_name, start=start, end=end)

df = pd.DataFrame(
    list(zip(cal.opens, cal.closes, cal.break_starts, cal.break_ends)),
    columns=["open", "close", "break_start", "break_end"],
    index=cal.closes.index,
)

print(f"Writing test CSV file to {path}")
df.to_csv(path, date_format="%Y-%m-%dT%H:%M:%SZ")
